!+
! Module: eexpmap_mod
!
! Authors:  Hermann Brunner - AIP
!           Alexis Finoguenov - AIP          
!          
! Description: EPIC exposure maps for use with the EPIC detection tasks
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module eexpmap_mod

  use cal
  use oal
  use dal
  use param  
  use dss
  use edetect_io_exp
  use pix2map_mod
  use errorHandling
!  use attcalc_mod
!  use emutils_mod
 
  implicit none 
     
  public eexpmap

  real(kind=double), dimension(:,:), allocatable :: fexpext
  type :: start_stopT
    real(kind=double), dimension(:), pointer :: start, stop
  end type


contains 

  !---------------------------------------------------------------------
  subroutine eexpmap
 !---------------------------------------------------------------------

    real(kind=single), dimension (:), allocatable :: amplo, amphi
    real(kind=double), dimension(:), pointer :: theta, phi
!!!!    real(kind=single), dimension(:), pointer :: theta, phi
    real(kind=double) :: dtheta, dphi, eps, ctime, hducut
    !real(kind=double), parameter :: pi = 3.141592654, pix = 4.0, &
    !     flngth=7500.0, fac = pix/3600.0 * (pi/180.0) * flngth, &
    !                                factheta = pix/fac
    real(kind=double), parameter :: pi = 3.141592654, pix = 4.0 
    real(kind=double) :: flngth, fac
    real(kind=double), dimension(:), pointer :: time, gstart, gstop
    real(kind=double), dimension (:), allocatable :: E, area, yextra
    real(kind=single), dimension(:), allocatable :: transmission, &
                                     effarea_cen, effgrad, teffgrad  
    real(kind=single) :: expos
    real(kind=single), dimension(:,:), allocatable :: &
                                           transmission_save, area_save   
    integer(kind=int16), dimension(0:600) :: rawX, rawY
    real(kind=double), dimension(:), pointer :: ra0, de0, pa0
    integer(kind=int16), dimension(:), pointer :: bbrawx, bbrawy,bbye,bbox,bbofs
    integer(kind=int8),  dimension(:),  pointer :: bbnr 
    integer(kind=int32) :: i, k, ngti_ext, & 
                           naxis1, naxis2, junk, j, &
                           ncoras, irange, itfilt, itfilt1, ngti1

! extra working parameters. Alexis.
    integer(kind=int32) :: npx0, npx1, npy0, npy1, mpx0, mpx1, mpy0, mpy1 
    integer(kind=int32) :: iexp,jexp,iccd,nccd,nccd0,ic0,ic1
    integer(kind=int16) :: ipx,ipy,ipyp,ipym2
    integer(kind=int16), parameter :: ipmin = 10
    integer(kind=int32) :: ngrade, igrade, ntgrade, maxgrad, nsubmode, nfilter, nband
    integer(kind=int32), dimension(0:100) :: grade, tgrade
!    real(kind=single), dimension(0:100) :: effg
    logical :: UserPatterns, useDetector, notsaid,  rndmz, useImageCS, &
               usedss,  qbb, qofs, withVig, fastmode, found, useoffset
! extra input switches
    real(kind=single) :: detxpx2mm,detypx2mm,detxcrpx, detycrpx,&
         & detxcrvl, detycrvl
    real(kind=single) :: x0,y0,x1,y1,xsp,xsn,ysp,ysn,fhx,fhy,fh2,wgh
    real(kind=single) :: xn,xm,xp,yn,ym,yp,rx,ry,rx0,ry0,pixoff,pixfac
    real(kind=single) :: wmm,wpp,wmp,wnp,wnm,wnn,wmn,wpn,wpm
    integer :: nsplit, lx, ly
    integer, dimension(:), pointer :: shape
!    integer(kind=int8), dimension(:,:,:), pointer :: patterns
    real(kind=double), dimension(:), pointer :: Xmm, Ymm, Zmm
!!!!    real(kind=single), dimension(:), pointer :: Xmm, Ymm, Zmm
    real(kind=single), dimension(0:3) :: Xmm1, Ymm1, Zmm1
    integer :: ix,iy
!    real(kind=single), dimension(:), pointer :: XmmSc, YmmSc, ZmmSc
    character(len=20) :: attinstr,name
    character(len=6) :: bbname, ocname
!   merge_gti characrer(len=12) :: ctype1, ctype2
    character(len=1024) :: stattit, gtiblock
    integer(kind=int32) :: diff0order, ipxd2, bbrawy0, bbrawy1, nbb0, nbb1
    real(kind=double), dimension(:,:), pointer :: Cmatrix
    real(kind=single), dimension(:,:,:), allocatable :: exp0
    real(kind=single), dimension(:,:,:,:), allocatable :: exp1
    real(kind=double), dimension(0:5000) :: fstart, fstop, tempstart, tempstop
    integer(kind=int32), dimension(0:100) :: iccd0, iccd1

    real(kind=double), parameter  :: radian = 57.29577951E0    
    real(kind=double), dimension(0:0), target :: tstart, tstop
    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsize
    integer(kind=int32) :: natt, lw, iblock, ntgti, ntempgti
    integer(kind=int32), dimension(12) :: winx0, winy0, windx, windy
    integer(kind=int32), parameter :: nattmax=1000
    integer(kind=int32), parameter :: nextmax=12
    real(kind=double), dimension(nattmax) :: xnew, ynew, crnew, srnew
    integer(kind=int32), dimension(12) :: ngti, acomp
    real(kind=single) :: xsky, ysky
    real(kind=single), dimension(0:5) :: xsky4, ysky4
    real(kind=single) :: fk, bix, biy
    real(kind=single), dimension(:), allocatable :: wght 
    real(kind=double) :: bmtrx10, bmtrx11, bmtrx20, bmtrx21
    integer(kind=int16), dimension(:,:), pointer :: map
    character(len=7) :: datamode, filterid
    character(len=20) :: submode                          
    character(len=255) :: wstr
! Pixel size parameters from EPN_LINCOORD_0001.CCF
    ! real(kind=single), parameter :: MM_PX_X=0.15, MM_PX_Y=0.15, MM_PX_LY=0.2
    ! real(kind=single), parameter :: MOS_PX=0.04
    real(kind=single) :: MM_PX_X, MM_PX_Y, MM_PX_LY, MOS_PX
    real(kind=single) :: xhsz_mm, yhsz_mm, xexpu, xexpl, yexpu, yexpl, fracx,&
         fracy, frac,rbbc,xshift,yshift,xcen,ycen,exptest
    integer(kind=int16) :: imod
    integer(kind=int16) :: iexp0,iexp1,jexp0,jexp1
    integer :: icomp, ncomp, ifilt, nfilt, ftype, ntest
    character(len=80) :: fname,binType

    integer(kind=int32) :: ibb, nbb,nofs,iofs,ibbc,ipxad,ipyad,ibbc1,irpx2,irbbc2
    logical :: pos, mgti, hasgti, hasgtin, close, closeos, hasoffset0 
    logical, dimension(nextmax) :: relgti

    logical, dimension(:), allocatable :: qalpha, q45
    real(kind=single), dimension(:), allocatable :: alsin, alcos, hs
    real(kind=double) :: alpha, dx, dy

    type(start_stopT), dimension(nextmax) :: gti
    type(start_stopT) :: tempgti
    type(DataSetT) :: set1,set2,set3
    type(DataSetT), dimension(:), allocatable :: setO
    type(ArrayT) :: arr0
    type(ArrayT), dimension(:), allocatable :: arr
    type(TableT) :: tabcoras,tabgti,taboffs
    type(TableT), dimension(nextmax) :: tabi
    type(ColumnT) :: coltime,coldeo,colrao,colpa,crawx,crawy,cye, &
                     colgtista,colgtisto,cox,cofs,cnr
    type(ColumnT), dimension(nextmax) :: colstart,colstop       
    type(BlockT) :: dblock
    type(DssT) :: dss
    type(DScompT), dimension(0:100) :: dssc 
    type(DSfilterT), dimension(0:100,0:100) :: dssf 
    type(RangeT), dimension(:), pointer :: frange
  

    !type :: ExpoT
    !  real(kind=single), dimension(:,:), pointer :: exp
    !end type
    !type(ExpoT), dimension(:), allocatable :: earr
    
    flngth = CAL_getMiscellaneousDataValue('FOCAL_LENGTH')
    call CAL_setState(EPN);
    MM_PX_X = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')
    MM_PX_Y = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_Y')
    MM_PX_LY = CAL_getMiscellaneousDataValue('LINCOORD:LINCOORD%MM_PX_LY')
    call CAL_setState(EMOS1);
    MOS_PX = CAL_getMiscellaneousDataValue('MM_PER_PIXEL_X')

    fac = pix/3600.0 * (pi/180.0) * flngth

! these will be read from user
    UserPatterns =  booleanParameter("withpatterns")
    useDetector =  booleanParameter("withdetcoords")
    useDSS = booleanParameter("usedss")
    withVig = booleanParameter("withvignetting")
    fastmode = booleanParameter("usefastpixelization")    
    useoffset = booleanParameter("useoffset")    
    if(useoffset) then
       hducut=realParameter("hducut")
    endif

    useImageCS = .true.
    rndmz = .false.
    diff0order = 0
    eps = realParameter("attrebin") / 3600.0
    ibbc = intParameter("badclean")

    nband=parameterCount("pimin")
    allocate (earr(0:nband-1))      
    allocate (setO(0:nband-1))
    allocate (arr(0:nband-1))
    allocate (e(0:nband-1))
    allocate (area(0:nband-1))
    allocate (transmission(0:nband-1))
    allocate (effgrad(0:nband-1))
    allocate (teffgrad(0:nband-1))
    allocate (amplo(0:nband-1))
    allocate (amphi(0:nband-1))
    allocate (effarea_cen(0:nband-1))
    allocate (area_save(0:300,0:nband))
    allocate (transmission_save(0:300,0:nband))
    allocate (wght(0:nband-1))
    allocate (yextra(0:700))

    do i=0,nband-1 
      amplo(i) = intParameter("pimin",i)
      amphi(i) = intParameter("pimax",i)
    end do
    
! input of attitude history file

    stattit=stringParameter("attitudeset")
    set2=dataset(stattit,Read)
    tabcoras = table(set2,"ATTHK")
    ncoras = numberOfRows(tabcoras)
    coltime = column(tabcoras,"TIME", Read)    
    time => real64Data(coltime)
    colrao = column(tabcoras,"AHFRA", Read)
    ra0 => real64Data(colrao)
    coldeo = column(tabcoras,"AHFDEC", Read)
    de0 => real64Data(coldeo)
    colpa = column(tabcoras,"AHFPA", Read)
    pa0 => real64Data(colpa)

! input of EPIC FITS image
    set1=dataset(stringParameter("imageset"),Read)
    arr0 = array(set1, "PRIMARY", Read)
    if (hasAttribute(set1,"INSTRUME")) then  
      attinstr  = stringAttribute(set1,"INSTRUME")
    else
      attinstr  = stringAttribute(arr0,"INSTRUME")
    end if      
    ngti = 0
    datamode = stringAttribute(set1,"DATAMODE")
    if (datamode /= "IMAGING") then
      ! .and. datamode /= "TIMING" .and. datamode /= "BURST") then
       write (errstr,*) "Mode ( ",datamode," ) not supported"
       call error("badDatamode")
    endif
    write(errstr,*) "Datamode: ",Datamode
    call message(AppMsg,Verbose)
    filterid = stringAttribute(set1,"FILTER")
    if (attinstr == "EPN") then
      if (index(filterid,"OPEN") > 0 .or. &
          index(filterid,"NONE") > 0) then
        nfilter=0 ! 12
      else if (index(filterid,"CLOSED") > 0 .or. &
               index(filterid,"Closed") > 0 .or. &
               index(filterid,"closed") > 0) then
        nfilter=1 ! 13 
      else if (index(filterid,"THIN1") > 0 .or. &
               index(filterid,"Thin1") > 0 .or. &
               index(filterid,"thin1") > 0) then
        nfilter=2 ! 14 
      else if (index(filterid,"THIN2") > 0 .or. &
               index(filterid,"Thin2") > 0 .or. &
               index(filterid,"thin2") > 0) then
        nfilter=3 ! 15 
      else if (index(filterid,"MEDIUM") > 0 .or. &
               index(filterid,"Medium") > 0 .or. &
               index(filterid,"medium") > 0 ) then
        nfilter=4 ! 16 
      else if (index(filterid,"THICK") > 0 .or. &
               index(filterid,"Thick") > 0 .or. &
               index(filterid,"thick") > 0) then
        nfilter=5 ! 17 
      else 
        nfilter=-1
      end if 
    end if 
    if (attinstr == "EMOS1" .or. attinstr == "EMOS2") then
      if (index(filterid,"OPEN") > 0 .or. &
          index(filterid,"NONE") > 0) then
        nfilter=0
      else if (index(filterid,"CLOSED") > 0 .or. &
               index(filterid,"Closed") > 0 .or. &
               index(filterid,"closed") > 0) then
        nfilter=1 
      else if (index(filterid,"THIN1") > 0 .or. &
               index(filterid,"Thin1") > 0 .or. &
               index(filterid,"thin1") > 0 ) then
        nfilter=2 
      else if (index(filterid,"THIN2") > 0 .or. &
               index(filterid,"Thin2") > 0 .or. &
               index(filterid,"thin2") > 0) then
        nfilter=3 
      else if (index(filterid,"MEDIUM") > 0 .or. &
               index(filterid,"Medium") > 0 .or. &
               index(filterid,"medium") > 0) then
        nfilter=4 
      else if (index(filterid,"THICK") > 0 .or. &
               index(filterid,"Thick") > 0 .or. &
               index(filterid,"thick") > 0) then
        nfilter=5 
      else 
        nfilter=-1
      end if 
    end if 
    if (nfilter < 0) then 
      call warning("NoFilt","No FILTER attribute found")
      nfilter=0
    end if
    if (hasAttribute(set1,"SUBMODE")) then  
      submode = stringAttribute(set1,"SUBMODE")
    else
      ! call warning("NoSubMode","No SUBMODE attribute found")
      submode = "PRIME_FULL_WINDOW"
    end if 
    if (attinstr == "EPN") then
      if (index(submode,"PRIME_FULL_WINDOW")  > 0 .or. &
          index(submode,"PrimeFullWindow")  > 0 ) then
        nsubmode=15
      else if (index(submode,"FULL_WINDOW_EXT") > 0 .or. &
               index(submode,"FullWindowExt") > 0)then
        nsubmode=16
      else if (index(submode,"PRIME_LARGE_WINDOW") > 0 .or. &
               index(submode,"PrimeLargeWindow") > 0) then
        nsubmode=17
      else if (index(submode,"PRIME_SMALL_WINDOW") > 0 .or. &
               index(submode,"PrimeSmallWindow") > 0) then
        nsubmode=18
      else if (index(submode,"PRIME_FULL_OFFSET")  > 0 .or. &
               index(submode,"PrimeFullOffset")  > 0) then
        nsubmode=21
      else if (index(submode,"PRIME_LARGE_OFFSET") > 0 .or. &
               index(submode,"PrimeLargeOffset") > 0) then
        nsubmode=22
      else if (index(submode,"OFFSET") > 0 .or. &
               index(submode,"Offset") > 0 ) then
        nsubmode=23
      else
        nsubmode=-1
      end if
    end if
    if (attinstr == "EMOS1" .or. attinstr == "EMOS2") nsubmode=-1 
    if (nsubmode == -1) then
      ! call warning ("NoSubMode","No valid SUBMODE attribute found")
      if (attinstr == "EPN") then 
        nsubmode = 15
      else
        nsubmode = 0
      end if
    end if  

    shape => dimensions(arr0)
    naxis1 = shape(0)-1
    naxis2 = shape(1)-1

    if (useImageCS) then

       if (hasAttribute(set1,"CTYPE1")) then
         binType=stringAttribute(set1,"CTYPE1")
       else
         binType=stringAttribute(arr0,"CTYPE1")
       endif
       if (useDetector) then
         if (trim(binType) /= "DETX" .and. trim(binType) /= "DETY") then
           call error("FileMismatch", &
                      "withdetcoords=yes, Coord. types DETX,DETY expected")  
         endif          
       else 
         if (trim(binType) /= "RA---TAN" .and. trim(binType) /= "DEC--TAN") then
           call error("FileMismatch", &
                      "withdetcoords=no, Coord. types RA---TAN,DEC--TAN expected")  
         endif  
       endif

       if (hasAttribute(set1,"CDELT1")) then
         detxpx2mm=real32Attribute(set1,"CDELT1")*flngth * (pi/180.0)
       else
         detxpx2mm=real32Attribute(arr0,"CDELT1")*flngth * (pi/180.0)
       end if     
       if (hasAttribute(set1,"CRPIX1")) then 
         detxcrpx=real32Attribute(set1,"CRPIX1")
       else
         detxcrpx=real32Attribute(arr0,"CRPIX1")
       end if
       if (hasAttribute(set1,"CRPIX2")) then
         detycrpx=real32Attribute(set1,"CRPIX2")
       else
         detycrpx=real32Attribute(arr0,"CRPIX2")
       end if
       if (hasAttribute(set1,"CRVAL1")) then  
         detxcrvl=real32Attribute(set1,"CRVAL1")
       else
         detxcrvl=real32Attribute(arr0,"CRVAL1")
       end if
       if (hasAttribute(set1,"CRVAL2")) then
         detycrvl=real32Attribute(set1,"CRVAL2")
       else
         detycrvl=real32Attribute(arr0,"CRVAL2")
       end if
       if (hasAttribute(set1,"CDELT2")) then
         detypx2mm=real32Attribute(set1,"CDELT2")*flngth * (pi/180.0)
       else
         detypx2mm=real32Attribute(arr0,"CDELT2")*flngth * (pi/180.0)
       end if
       if (useDetector) then
         detxcrvl=detxcrvl*flngth*pi/(180.0*72000.0)
         detycrvl=detycrvl*flngth*pi/(180.0*72000.0)
         detxpx2mm=detxpx2mm/72000.0
         detypx2mm=detypx2mm/72000.0
       end if
    else
! this is only for development purposes, not for the user
       detxpx2mm=fac
       detxcrpx=naxis1/2.
       detycrpx=naxis2/2.
       detxcrvl=0.
       detycrvl=0.
       detypx2mm=fac
    end if

    ranomrad = detxcrvl/radian
    sdecnom = dsin(detycrvl/radian)
    cdecnom = dcos(detycrvl/radian)
    skypixsize = abs(detypx2mm) / flngth * (180.0/pi)
   
! Event Patterns for QE

    if(.not. UserPatterns) then
      ngrade = 0
      grade(1)=0
    else
       wstr = stringParameter("pattern")
       lw= len_trim(wstr)
       ngrade=0
       i=1
       do while (i <= lw)
          junk=i
          do j=junk,junk
             if(wstr(j:j) == ' ') exit
             ngrade=ngrade+1
             read(wstr(j:min((j+1), lw)),*)grade(ngrade)
             i=i+1
          end do
          i=i+1
       end do
       if (attinstr=="EPN") maxgrad=1
       if (attinstr(1:4)=="EMOS") maxgrad=31
       ntgrade=0
       found=.false.
       do i=0,maxgrad
         do j=1,ngrade
           if (grade(j)==i) found=.true.
         end do
         if (.not.found) then
           ntgrade=ntgrade+1
           tgrade(ntgrade)=i  
         end if
       end do 
    end if
    write(errstr,*) 'Number of Grades: ',ngrade
    call message(AppMsg,VERBOSE)
    write(errstr,*) 'Grades: ',(grade(i),i=1,ngrade)
    call message(AppMsg,VERBOSE)
    if(ngrade == 0) then
      write(errstr,*) 'No grade selection. QE will be set to 1.'
      call message(AppMsg,VERBOSE)
    end if

!    if (attinstr == "EPN" .and. ngti_ext /= EPN_CCD_LAST) call warning&
!         ("NumGTI", "Number of GTI extensions /= number of chips") 
!    if (attinstr(1:4) == "EMOS" .and. ngti_ext /= EMOS_CCD_LAST) call warning&
!         ("NumGTI", "Number of GTI extensions /= number of chips") 

    if (attinstr == "EPN") then
      ngti_ext = EPN_CCD_LAST
    else
      ngti_ext = EMOS_CCD_LAST
    end if 

    relgti=.false.
    ngti=0
        

    if (useDSS) then

    allocate (tempgti%start(0:0)) 
    allocate (tempgti%stop(0:0))  
    dblock = block(arr0)
    dss = dataSubspace(dblock)
    ncomp = dssComponents(dss)
    acomp=-1
    do icomp = 0, ncomp-1
      dssc(icomp) = dssComponent(dss,icomp)
      nfilt = dssFilters(dssc(icomp))
      iccd0=-1
      iccd1=99  
      iccd=0
      irange=0
      itfilt=0  
      ntempgti=0  

      deallocate (tempgti%start)
      deallocate (tempgti%stop)
      allocate (tempgti%start(0:0)) 
      allocate (tempgti%stop(0:0))        

      do ifilt = 0,nfilt-1
        dssf(ifilt,icomp) = dssFilter(dssc(icomp),ifilt) 
        ftype = dssFiltertype(dssf(ifilt,icomp))
        if (ftype == rangefiltert) then
          fname = trim(dssFilterName(dssf(ifilt,icomp)))
          frange => dssRanges(dssf(ifilt,icomp))
          if (fname == "CCDNR") then
            do i=0,size(frange)-1
              if (frange(i)%lowerB /= Undef) iccd0(iccd) = int(frange(i)%lower)
              if (frange(i)%upperB /= Undef) iccd1(iccd) = int(frange(i)%upper)
              iccd=iccd+1
            end do
          end if
          
          if (fname == "TIME") then
            do i=0,size(frange)-1
              if (frange(i)%lowerB /= Undef) then
                fstart(irange)=frange(i)%lower
              else
                fstart(irange)=0d0
              end if
              if (frange(i)%upperB /= Undef) then
                fstop(irange)=frange(i)%upper
              else
                fstop(irange)=1d30
              end if
              irange=irange+1
              if (irange > 5000) call error("ArrayOutOfRange", &
                 "DSS contains more than 5000 time intervals")
            end do 
            itfilt=itfilt+1
          end if
        end if 
        if (ftype == gtifiltert) then
          fname = trim(dssFilterName(dssf(ifilt,icomp)))
            gtiblock = dssFilterBlock(dssf(ifilt,icomp))              
            iblock = index(gtiblock,":STDGTI")            
            if (iblock > 0) gtiblock = trim(gtiblock(iblock+1:))
            if (hasBlock(set1,gtiblock)) then
              tabgti = table(set1,gtiblock)
              ntgti = numberOfRows(tabgti) 
              colgtista = column(tabgti,"START", Read)    
              gstart => real64Data(colgtista)
              colgtisto = column(tabgti,"STOP", Read)    
              gstop => real64Data(colgtisto)
              do i=0,ntgti-1
                fstart(irange)=gstart(i)
                fstop(irange)=gstop(i)
                irange = irange+1
                if (irange > 5000) call error("ArrayOutOfRange", &
                   "DSS contains more than 5000 time intervals")
              end do
            end if
            itfilt=itfilt+1
            call release (colgtista)
            call release (colgtisto)
            call release (tabgti)
        end if  
      end do ! ifilt
 
       do i=1,ngti_ext
         mgti=.true.
         do j=0,iccd-1
           ic0 = iccd0(j)
           ic1 = iccd1(j)
           !if (.not. (i>ic0.and.i<ic1 .or. i==ic0.and.i==ic1)) merge=.false.
           !if ((i<=ic0.or.i>=ic1).and.(i/=ic0.or.i/=ic1)) merge=.false.
           if ((i<=ic0.or.i>=ic1).and.(i/=ic0.or.i/=ic1)) mgti=.false.
         end do
         if (mgti) then
            !combine filter gtis with &&              
           itfilt1=itfilt+ntempgti-1
           call merge_gti(tempgti, ntempgti, fstart, fstop, irange, itfilt, itfilt1)
            !combine component gtis with ||
           itfilt=ntempgti
           itfilt1=0
           do j=0,ntempgti-1
             tempstart(j)=tempgti%start(j)
             tempstop(j)=tempgti%stop(j)
           end do
           irange=ntempgti
           call merge_gti(gti(i), ngti(i), tempstart, tempstop, irange, itfilt, itfilt1)
           acomp(i)=icomp  
         end if   
       end do        
         
    end do
    
    ntest=0
    do i=1,ngti_ext
      ntest=ntest+ngti(i)
    end do 
    if (ntest==0) call error ("NoGTI", &
                  "No GTI or TIME filter in DSS")

    else ! now part for usedss=.false.

! GTI extensions with the names STDGTI, STDGTI1 .. STDGTI9, 
! STDGTI01 .. STDGTI12 are accepted
    hasgti=.false.
    name = "STDGTI"
    if (hasBlock(set1,name)) then
      hasgti=.true.
      tabi(1) = table(set1,name)  
      ngti1 = numberOfRows(tabi(1))
      colstart(1) = column(tabi(1),"START", Read)    
      gti(1)%start => real64Data(colstart(1))
      colstop(1) = column(tabi(1),"STOP", Read)    
      gti(1)%stop => real64Data(colstop(1))
      relgti(1)=.true.
!      ngti_ext = 1
    else
      name = "GTI" ! to work with Xselect output
      if (hasBlock(set1,name)) then
         hasgti=.true. 
         tabi(1) = table(set1,name)  
         ngti1 = numberOfRows(tabi(1))
         colstart(1) = column(tabi(1),"START", Read)    
         gti(1)%start => real64Data(colstart(1))
         colstop(1) = column(tabi(1),"STOP", Read)    
         gti(1)%stop => real64Data(colstop(1))
         relgti(1)=.true. 
!         ngti_ext = 1
      end if
    end if
    hasgtin=.false.
    do i = 1,nextmax
      write(name,"(a,i2.2)") "STDGTI", i  
      if (hasBlock(set1,name)) then
        hasgti=.true.
        hasgtin=.true. 
        if (relgti(i)) then
          call release(colstart(i))
          call release(colstop(i))
          call release(tabi(i))
        end if
        tabi(i) = table(set1,name)
        ngti(i) = numberOfRows(tabi(i))
        colstart(i) = column(tabi(i),"START", Read)    
        gti(i)%start => real64Data(colstart(i))
        colstop(i) = column(tabi(i),"STOP", Read)    
        gti(i)%stop => real64Data(colstop(i))
        relgti(i)=.true.
!        ngti_ext = i
      end if
      if (i < 11) then
        write(name,"(a,i1)") "STDGTI", i   
        if (hasBlock(set1,name)) then
          hasgti=.true.
          hasgtin=.true. 
          tabi(i) = table(set1,name)
          ngti(i) = numberOfRows(tabi(i))
          colstart(i) = column(tabi(i),"START", Read)    
          gti(i)%start => real64Data(colstart(i))
          colstop(i) = column(tabi(i),"STOP", Read)    
          gti(i)%stop => real64Data(colstop(i))
          relgti(i)=.true.
!          ngti_ext = i
        end if
      end if
    end do

    if (.not. hasgti) then

     call warning ("noGTI","No GTI extension found in input image") 

! start and stop are set to moderately meaningful values in case 
! no GTI extensions are present. 
    if (hasAttribute(set1,"EXPOSURE")) then
      expos = real32Attribute(set1,"EXPOSURE")
! fix to work with early data. Alexis
!     if(expos .gt. 1.e5) expos=1.
!!!      ngti(1) = 1
      tstart(0) = time(0)
      tstop(0) = time(0) + expos
      do i=1,ngti_ext 
        gti(i)%start => tstart
        gti(i)%stop => tstop
        ngti(i)=1    
      end do
      ngti1=1
!      ngti_ext = 0
    else if (hasAttribute(arr0,"EXPOSURE")) then
      expos = real32Attribute(arr0,"EXPOSURE")
! fix to work with early data
!     if(expos .gt. 1.e5) expos=1.
!!!      ngti(1) = 1
      tstart(0) = time(0)
      tstop(0) = time(0) + expos
      do i=1,ngti_ext
        gti(i)%start => tstart
        gti(i)%stop => tstop
        ngti(i)=1      
      end do
      ngti1=1
!      ngti_ext = 0
    else
      call error ("onGTI/noexposure", &
                  "No GTI and no exposure available")
    end if 

    end if

    if (.not.hasgtin) then
      ngti(1)=ngti1
      do i=2, ngti_ext
        allocate (gti(i)%start(0:ngti(1)-1))
        allocate (gti(i)%stop(0:ngti(1)-1))
        gti(i)%start = gti(1)%start
        gti(i)%stop = gti(1)%stop
        ngti(i)=ngti(1) 
       end do
     end if
      
!    ngti_ext0 = ngti_ext 
!    if (ngti_ext > 0) then
!      ngti_ext0 = ngti_ext
!    else
!      ngti_ext0 = 0
!      ngti_ext = 1
!      ngti(1) = 1
!    end if

    end if ! usedss

    if(attinstr .ne. 'EPN') then
       nccd0=EMOS_CCD_FIRST
       nccd=EMOS_CCD_LAST
    else
       nccd0=EPN_CCD_FIRST
       nccd=EPN_CCD_LAST
    end if

! prepare to read the exposure extension and bad pixel stuff
    set3=dataset(stringParameter("eventset"),Read)
    bbname = "BADPIX"
    ocname = "OFFSET"

! create exposure images
    do i=0,nband-1
      setO(i)=dataset(stringParameter("expimageset",i),Create)
      arr(i)=addArray(setO(i),"PRIMARY",Real32,shape)
      earr(i)%exp => real32Array2Data(arr(i))          
      earr(i)%exp = 0.
    end do
    yextra=0.

! filterId needs to be read from input image
    if (attinstr.eq.'EPN') then
       call CAL_SetState(instrument=EPN, &
            filterId=nfilter, ccdModeId=nsubmode, randomize=rndmz)
       call OAL_SetState(instrument=EPN)
       imod = 5
       nsplit=max(4,ceiling(0.5*MM_PX_X/detxpx2mm),ceiling(0.5*MM_PX_Y/detypx2mm))
       xhsz_mm= 0.5* MM_PX_X
       yhsz_mm= 0.5* MM_PX_Y
!       if(ipy .eq. EPN_MAX_Y) yhsz_mm= 0.5* MM_PX_LY
       yextra(EPN_MAX_Y) = 0.5* MM_PX_LY - yhsz_mm
       xshift=0.
       yshift=0.
    end if

    if (attinstr.eq.'EMOS1') then
       call CAL_SetState(instrument=EMOS1, &
            filterId=nfilter,ccdNodeId=0,ccdModeId=nsubmode, randomize&
            & =.false.)
       call OAL_SetState(instrument=EMOS1)
       imod = 20
       nsplit=max(1,ceiling(0.5*MOS_PX/detxpx2mm),ceiling(0.5*MOS_PX/detypx2mm))
       xhsz_mm= 0.5* MOS_PX
       yhsz_mm= 0.5* MOS_PX
       xshift=0.
       yshift=0.
    end if
       
    if (attinstr.eq.'EMOS2') then
       call CAL_SetState(instrument=EMOS2, &
            filterId=nfilter,ccdNodeId=0,ccdModeId=nsubmode, randomize&
            & =.false.)
       call OAL_SetState(instrument=EMOS2)
       imod = 20
       nsplit=max(1,ceiling(0.5*MOS_PX/detxpx2mm),ceiling(0.5*MOS_PX/detypx2mm))
       xhsz_mm= 0.5* MOS_PX
       yhsz_mm= 0.5* MOS_PX
       xshift=0.
       yshift=0.
    end if

    pixoff=2.0*xhsz_mm/abs(detxpx2mm)*(0.5/float(nsplit)-0.5)
    pixfac=2.0*xhsz_mm/abs(detxpx2mm)/float(nsplit)
    fhx=xhsz_mm/(abs(detxpx2mm)*nsplit)
    fhy=yhsz_mm/(abs(detypx2mm)*nsplit)

    fh2=4.0*fhx*fhy

! we were advised to use medium quality 
    call CAL_setState(accuracyLevel = ACCURACY_MEDIUM)

!    if (.not. useDetector) then
! precalculation of the detector dead times using the GTI. For sky
    ! coordinates, however, this has to be done on pointing histogram time
    ! step. 

       call atthst (ranomrad, sdecnom, cdecnom, skypixsize, time, ra0, de0,&
            pa0, ncoras, gti, ngti, ngti_ext, nccd, set3,  & ! input 
            xnew, ynew, crnew, srnew, natt, eps, &
            winx0, winy0, windx, windy, nextmax)        
!       print*,xnew(1),ynew(1)
!    end if

! Here we allocate the memory for pixel transformation arrays
    allocate (qalpha(1:natt))
    allocate (alsin(1:natt))
    allocate (alcos(1:natt))
    allocate (hs(1:natt))
    allocate (q45(1:natt))
    do k=1,natt
       qalpha(k) = .true.
    enddo
 
! 80 is to suppress errors in attcalc 
    if(.not.fastmode) skypixsize = skypixsize/80. 

    if (fastmode.and.natt.gt.18) then
      fastmode=.false.
      skypixsize = skypixsize/80.
      call warning("NoFastPixelization", &
      "Too many attitude bins; fast pixelization turned off")
    end if

    if (fastmode) then
      allocate (exp0(1:natt,0:naxis1-1,0:naxis2-1)) 
      exp0(1:natt,0:naxis1-1,0:naxis2-1)=0.0   
      allocate (exp1(0:nband-1,1:natt,0:naxis1-1,0:naxis2-1)) 
      exp1(0:nband-1,1:natt,0:naxis1-1,0:naxis2-1)=0.0   
    end if

    if (useDetector) then
      do iccd=1,nccd
        do k=2,natt
          fexpext(iccd,1)=fexpext(iccd,1)+fexpext(iccd,k)
        end do
      end do
      natt=1 
    end if 

    do i = 1, nextmax
      if (relgti(i)) then
        call release(colstart(i)) 
        call release(colstop(i))
        call release(tabi(i))
      end if
    end do
    call release(colrao)
    call release(coldeo)
    call release(colpa)
    call release(coltime)
    call release(tabcoras)
   
!    call CAL_piToeV(amplo,emin)
!    call CAL_piToeV(amphi,emax)
!    e(0) = (emin(0)+emax(0))/2.0
    
     do i=0,nband-1
       e(i) = (amplo(i)+amphi(i))/2.0
       if (withVig) then
         effarea_cen(i) = CAL_getEffectiveArea(e(i),0d0,0d0) * &
                          CAL_getFilterTransmission(e(i),0d0,0d0)
       else
         effarea_cen = 1.0
       end if
     end do

! effective area at optical axis

    !------------------------------------------!
    !   get  CC2->SC Cmatrix and angles        !
    !------------------------------------------!
    do i=nextmax,-1,1
      if (ngti(i)>0) ctime=gti(i)%start(0)
    end do 
    call CAL_getCamCoord2ToSacCoord(ctime, Cmatrix)
    bmtrx10 = Cmatrix(1,0)
    bmtrx11 = Cmatrix(1,1)
    bmtrx20 = Cmatrix(2,0)
    bmtrx21 = Cmatrix(2,1)
!*     write (*,*) Cmatrix(0,0),Cmatrix(0,1),Cmatrix(0,2)
!*     write (*,*) Cmatrix(1,0),Cmatrix(1,1),Cmatrix(1,2)
!*     write (*,*) Cmatrix(2,0),Cmatrix(2,1),Cmatrix(2,2)
    call CAL_releaseMemory(Cmatrix)

! read OFFSETS extension:
    hasoffset0=.false.
    if (attinstr .eq. 'EPN' .and. useoffset) then
      if (hasBlock(set3,"OFFSETS")) then
        taboffs = table(set3,"OFFSETS")
        nofs = numberOfRows(taboffs) 
        if (nofs > 0) then
            cox = column(taboffs,"RAWX", Read)    
            bbox => int16Data(cox)
            cofs = column(taboffs,"OFFSETX", Read)
            bbofs => int16Data(cofs)
            cnr  = column(taboffs,"CCDNR", Read)
            bbnr  => int8Data(cnr)
            hasoffset0=.true.
            write(errstr,*) 'read offsets for ',nofs, ' columns'
            call message(AppMsg,VERBOSE)
        endif               
      endif
    endif

!loop over ccds.
    close=.false.
    closeos=.false.
    do iccd=nccd0, nccd
!!!*   do iccd=nccd0,nccd0 ! for quick test
      if (ngti(iccd) > 0) then 
       write(errstr,*) 'doing ',attinstr,' ccd chip #',iccd
       call message(AppMsg,VERBOSE)
       call CAL_setState(ccdChipId=iccd)
       call CAL_getFOVmap(map)
       npx0=EPN_MIN_X
       npx1=EPN_MAX_X
       npy0=EPN_MIN_Y
       npy1=EPN_MAX_Y
       mpx0=max(winx0(iccd),EPN_MIN_X)
       mpx1=min(mpx0+windx(iccd)-1,EPN_MAX_X)
       mpy0=max(winy0(iccd),EPN_MIN_Y)
       mpy1=min(mpy0+windy(iccd)-1,EPN_MAX_Y)
!!!       mpy0=31
!!!       mpy1=200
       if(attinstr .ne. 'EPN') then
         npx0=EMOS_MIN_X
         npx1=EMOS_MAX_X
         npy0=EMOS_MIN_Y
         npy1=EMOS_MAX_Y
         mpx0=max(winx0(iccd),EMOS_MIN_X)
         mpx1=min(mpx0+windx(iccd)-1,EMOS_MAX_X)
         mpy0=max(winy0(iccd),EMOS_MIN_Y)
         mpy1=min(mpy0+windy(iccd)-1,EMOS_MAX_Y)
       end if
! Read Bad pixel extension
!       if(nbb > 0 .and. iccd > 1) then
        if (close) then 
! clean previous extension
          call release (crawx)
          call release (crawy)
          call release (cye)
          call release (tabcoras)
       end if
        if (closeos) then 
! clean previous extension
          call release (cox)
          call release (cofs)
          call release (taboffs)
          closeos=.false.
       end if
       write(name,"(i2.2)") iccd  
       name=bbname//name(1:2)
       if (hasBlock(set3,name)) then
          tabcoras = table(set3,name)
          nbb = numberOfRows(tabcoras)
          crawx = column(tabcoras,"RAWX", Read)    
          bbrawx => int16Data(crawx)
          crawy = column(tabcoras,"RAWY", Read)
          bbrawy => int16Data(crawy)
          cye = column(tabcoras,"YEXTENT", Read)
          bbye => int16Data(cye)
          close=.true.
       else
          nbb=0
          call warning("NoBadPix","BadPixel extension not found")
       end if  

! get the information about the offset columns
       if (attinstr .eq. 'EPN' .and. useoffset .and. .not. hasoffset0) then
          write(name,"(i2.2)") iccd  
          name=ocname//name(1:2)
          if (hasBlock(set3,name)) then
             taboffs = table(set3,name)
             nofs = numberOfRows(taboffs)
             if (nofs > 0 .and. hasColumn(taboffs,"RAWX")) then
                cox = column(taboffs,"RAWX", Read)    
                bbox => int16Data(cox)
                cofs = column(taboffs,"OFFSET", Read)
                bbofs => int16Data(cofs)
                closeos=.true.
                write(errstr,*) 'read offsets in a number of',nofs
                call message(AppMsg,VERBOSE)
             endif
          else
             closeos=.false.
             call warning("NoOffset","EPN Offset extension not found")
          end if
       end if
               
       area_save=-1.0  

! get bad pixel identificator
       rbbc=get_BadRadius(ibbc)
       ibbc1=rbbc
       irbbc2=rbbc**2

! loop over the pixels, PIXCOORD1
       do ipx=npx0, npx1 
          qofs=.false.
          if (modulo(ipx,imod) == 0) area_save=-1.0

          do ipy=npy0,npy1 
            rawX(ipy)=ipx
            rawY(ipy)=ipy
          end do 

          nbb0=nbb+1
          nbb1=0
          do ibb=0,nbb-1
            if (bbrawx(ibb) >= ipx-ibbc1 .and. &
                bbrawx(ibb) <= ipx+ibbc1) then
              nbb0=min(nbb0,ibb)
              nbb1=max(nbb1,ibb)
            end if
          end do

! EPN offset columns
          if(hasoffset0) then
             do ipxad=ipx-ibbc1,ipx+ibbc1
                do iofs=0,nofs-1
                   if(ipxad == bbox(iofs) .and. bbofs(iofs) > hducut .and. bbnr(iofs) == iccd) then
                      write(errstr,*) 'offset column ',  bbox(iofs),' will be excluded'
                      call message(AppMsg,VERBOSE)
                      qofs =.true.
                   endif
                end do
             end do
          endif
          if(closeos) then
             do ipxad=ipx-ibbc1,ipx+ibbc1
                do iofs=0,nofs-1
                   if(ipxad == bbox(iofs) .and. bbofs(iofs) > hducut) then
                      write(errstr,*) 'offset column ',  bbox(iofs),' will be excluded'
                      call message(AppMsg,VERBOSE)
                      qofs =.true.
                   endif
                end do
             end do
          endif

          call CAL_rawXY2mm(rawX(npy0:npy1),rawY(npy0:npy1),Xmm,Ymm,Zmm,2)
          call CAL_camCoord2ToTelCoord(Xmm(npy0-1:npy1-1),Ymm(npy0-1:npy1-1),&
                                       Zmm(npy0-1:npy1-1),theta, phi)
          notsaid=.true.
          do ipyp=npy0,npy1
             do ipy=ipyp,ipyp ! this is done to accomplish different exits
! check if this pixel is bad or out of FOV
!                if(CAL_isPixelBad(ipx,ipy,yextent,defect,uplinked)) exit
                qbb=.false.
                if(map(ipx-1,ipy-1) .ne. 1) qbb=.true.
                if (ipx<mpx0 .or. ipx>mpx1 .or. &
                    ipy<mpy0 .or. ipy>mpy1) qbb=.true. 

                do ibb=nbb0,nbb1
                  do ipxad=ipx-ibbc1,ipx+ibbc1
                    if (ipxad==bbrawx(ibb)) then  
                      ipxd2=(ipxad-ipx)**2
                      bbrawy0=bbrawy(ibb)
                      bbrawy1=bbrawy0+bbye(ibb) 
                      do ipyad=ipy-ibbc1,ipy+ibbc1
                        if (ipyad >= bbrawy0 .and. ipyad < bbrawy1) then
                            irpx2=ipxd2+(ipyad-ipy)**2
                            if(irpx2 .le. irbbc2) qbb =.true.
                        end if
                      end do
                    end if
                  end do
                end do
!               if(qbb) exit
! Borders 
                do ipxad=ipx-ibbc1,ipx+ibbc1
                   do ipyad=ipy-ibbc1,ipy+ibbc1
                      if(ipxad == npx0-1 .or. ipxad == npx1+1 .or.&
                           ipyad== npy0-1 .or. ipyad == npy1+1 )qbb =.true.
                   end do
                end do
!               if(qbb) exit
!             if (qsmpl) exit
                              
! transform PIXCOORD1 --> CAMCOORD2
                ! 2 stands for PIXCOORD1 --> CAMCOORD2 transformation

                if (area_save(ipy/imod,0)< 0) then
  
                  if (ngrade==0) then
                    effgrad=1.
                  else 
                    ipym2=ipy+imod/2
                    effgrad=0.
                    do igrade=1,ngrade                  
                      do i=0,nband-1
                        ! get quantum efficiency 
                        effgrad(i) = effgrad(i) + &
                        CAL_getQuantumEfficiency(e(i),ipx,ipym2,grade(igrade))
                      end do
                    end do
                    teffgrad=0.
                    do igrade=1,ntgrade                  
                      do i=0,nband-1
                        ! get quantum efficiency 
                        teffgrad(i) = teffgrad(i) + &
                        CAL_getQuantumEfficiency(e(i),ipx,ipym2,tgrade(igrade))
                      end do
                    end do
                    do i=0,nband-1
                      effgrad(i)=effgrad(i)/(effgrad(i)+teffgrad(i))
                    end do
                  end if

                  dtheta=theta(ipy-1)
                  dtheta=min(theta(ipy-1),999.0d0)  
!!!!!!                  dtheta=min(theta(ipy-1),999.0)  
                  dphi=phi(ipy-1)
                  if (dphi < 0d0) dphi = 2d0*pi + dphi
                  do i=0,nband-1
                    if (withVig) then
                      area(i)=CAL_getEffectiveArea(e(i),dtheta,dphi) * &
                              CAL_getFilterTransmission(e(i),dtheta,dphi) / &
                              effarea_cen(i) * effgrad(i)
                    else
                      area(i)=effgrad(i)
                    end if 
                    area_save(ipy/imod,i) = area(i)  
                  end do 
                else
                  do i=0,nband-1
                    area(i) = area_save(ipy/imod,i)
                  end do
                end if 
                
                if (qbb.or.qofs) area(0:nband-1)=0.0

                if (useDetector) then
                   ! account for the pixel size
!                    xexpu=(Xmm(ipy-1)+xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
!                    xexpl=(Xmm(ipy-1)-xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
!                    yexpu=(Ymm(ipy-1)+yhsz_mm - detycrvl)/detypx2mm + detycrpx
!                    yexpl=(Ymm(ipy-1)-yhsz_mm - detycrvl)/detypx2mm + detycrpx
                   xexpu=(Xmm(ipy-1) - detxcrvl)/detxpx2mm + detxcrpx
                   xexpl=(Xmm(ipy-1)-2*xhsz_mm - detxcrvl)/detxpx2mm + detxcrpx
                   yexpu=(Ymm(ipy-1) - detycrvl)/detypx2mm + detycrpx
                   yexpl=(Ymm(ipy-1)-2*(yhsz_mm+yextra(ipy)) - detycrvl)/detypx2mm + detycrpx
                   iexp0=xexpl
                   iexp1=xexpu
                   jexp0=yexpl
                   jexp1=yexpu
                   
!*                 if(xexpl .lt. 0) iexp0=iexp0-1
!*                 if(xexpu .lt. 0) iexp1=iexp1-1
!*                 if(yexpl .lt. 0) jexp0=jexp0-1
!*                 if(yexpu .lt. 0) jexp1=jexp1-1
! this is true, but not necessary, since we discard these values later

                   do iexp=iexp0,iexp1
                      do jexp=jexp0,jexp1
                         if(iexp .ge. 0 .and. iexp .le. naxis1 .and. &
                            jexp .ge. 0 .and. jexp .le. naxis2) then
                            fracx=min(xexpu,real(iexp+1))-max(xexpl,real(iexp))
                            fracy=min(yexpu,real(jexp+1))-max(yexpl,real(jexp))
                            frac=fracx*fracy ! /(xexpu-xexpl)**2 
                            ! we can't use Y here
                            ! since we want to account for different pixel
                            ! size in Y direction
                            do i=0,nband-1
                              earr(i)%exp(iexp,jexp)=earr(i)%exp(iexp,jexp) + &
                                                   frac*fexpext(iccd,1)*area(i)
                            end do
                         end if
                      end do
                   end do
                   if(notsaid) then
                      notsaid=.false.
                   end if
                   exit
                else

                 if (fastmode) then
 
                  rx0 = Xmm(ipy-1)/detypx2mm+pixoff
                  ry0 = Ymm(ipy-1)/detypx2mm+pixoff
                  do lx=0,nsplit-1
                    rx=rx0+lx*pixfac
                    do ly=0,nsplit-1
                      ry=ry0+ly*pixfac
          
                      bix = (bmtrx10 * rx) + (bmtrx11 * ry) 
                      biy = (bmtrx20 * rx) + (bmtrx21 * ry)

                      do k=1,natt
                       if (fexpext(iccd,k)>0.0) then
 
                        xsky = detxcrpx + xnew(k) &
                               - (crnew(k)*bix - srnew(k)*biy)
                        ysky = detycrpx + ynew(k) &
                               - (srnew(k)*bix + crnew(k)*biy)
!                       apply empiric offsets:
                        xsky=xsky-0.5                       
                        ysky=ysky-0.5
                        ix=xsky
                        iy=ysky
                        if (ix>0 .and. ix<naxis1-1 .and. &
                            iy>0 .and. iy<naxis2-1) then 
                          x0=ix
                          y0=iy
                          x1=ix+1 
                          y1=iy+1
                          xsp=xsky+fhx
                          xsn=xsky-fhx
                          ysp=ysky+fhy
                          ysn=ysky-fhy
                          if (xsn>x0 .and. xsp<x1 .and. &
                              ysn>y0 .and. ysp<y1) then 
                            do i=0,nband-1
                               wgh=area(i)*fexpext(iccd,k) 
                               if (wgh > 0.0) then
                                 exp1(i,k,ix,iy)=exp1(i,k,ix,iy)+fh2*wgh
                               end if
                            end do
                            exp0(k,ix,iy)=exp0(k,ix,iy)+fh2 
                          else
                            xm=min(xsp,x1)-max(xsn,x0)
                            ym=min(ysp,y1)-max(ysn,y0)
                            xp=max(xsp-x1,0.0)
                            yp=max(ysp-y1,0.0)
                            xn=max(x0-xsn,0.0)                    
                            yn=max(y0-ysn,0.0)
                            wmm=xm*ym
                            wpp=xp*yp
                            wmp=xm*yp
                            wnp=xn*yp
                            wnm=xn*ym
                            wnn=xn*yn         
                            wmn=xm*yn
                            wpn=xp*yn                          
                            wpm=xp*ym

                            do i=0,nband-1
                              wgh=area(i)*fexpext(iccd,k) 
                              if (wgh > 0.0) then
                                exp1(i,k,ix,iy)= &
                                exp1(i,k,ix,iy  )+wmm*wgh
                                exp1(i,k,ix-1,iy)= &
                                exp1(i,k,ix-1,iy)+wnm*wgh
                                exp1(i,k,ix+1,iy)= &
                                exp1(i,k,ix+1,iy)+wpm*wgh
                                if (yp > 0.0) then
                                  exp1(i,k,ix+1,iy+1)= &
                                  exp1(i,k,ix+1,iy+1)+wpp*wgh
                                  exp1(i,k,ix  ,iy+1)= &
                                  exp1(i,k,ix  ,iy+1)+wmp*wgh
                                  exp1(i,k,ix-1,iy+1)= &
                                  exp1(i,k,ix-1,iy+1)+wnp*wgh
                                end if
                                if (yn > 0.0) then
                                  exp1(i,k,ix-1,iy-1)= &
                                  exp1(i,k,ix-1,iy-1)+wnn*wgh
                                  exp1(i,k,ix  ,iy-1)= &
                                  exp1(i,k,ix  ,iy-1)+wmn*wgh
                                  exp1(i,k,ix+1,iy-1)= &
                                  exp1(i,k,ix+1,iy-1)+wpn*wgh
                                end if 
                              end if
                            end do
                            exp0(k,ix  ,iy  )=exp0(k,ix  ,iy  )+wmm
                            exp0(k,ix-1,iy  )=exp0(k,ix-1,iy  )+wnm
                            exp0(k,ix+1,iy  )=exp0(k,ix+1,iy  )+wpm
                            exp0(k,ix+1,iy+1)=exp0(k,ix+1,iy+1)+wpp
                            exp0(k,ix  ,iy+1)=exp0(k,ix  ,iy+1)+wmp
                            exp0(k,ix-1,iy+1)=exp0(k,ix-1,iy+1)+wnp
                            exp0(k,ix-1,iy-1)=exp0(k,ix-1,iy-1)+wnn
                            exp0(k,ix  ,iy-1)=exp0(k,ix  ,iy-1)+wmn
                            exp0(k,ix+1,iy-1)=exp0(k,ix+1,iy-1)+wpn
                          end if
                        end if
                       end if
                      end do ! k
                    end do
                  end do
     
                 else

                    Xmm1(0)=Xmm(ipy-1)+xhsz_mm
                    Ymm1(0)=Ymm(ipy-1)-(yhsz_mm+yextra(ipy))
                    Zmm1(0)=Zmm(ipy-1)
                    do i=2,4
                       Xmm1(i-1)=Xmm1(0)-(1+(-1)**((i+1)/2))*xhsz_mm
                       Ymm1(i-1)=Ymm1(0)+(1+(-1)**((i+2)/2))*(yhsz_mm+yextra(ipy))
                    end do
                    do k=1,natt
                     if (fexpext(iccd,k)>0.0) then
                       do i=1,4
                          rx = Xmm1(i-1)/ detypx2mm*80.
                          ry = Ymm1(i-1)/ detypx2mm*80.
                          bix = (bmtrx10 * rx) + (bmtrx11 * ry) ! = Ysc 
                          biy = (bmtrx20 * rx) + (bmtrx21 * ry) ! = Zsc
                          xsky4(i) = detxcrpx  + xnew(k)&
                               - (crnew(k)*bix - srnew(k)*biy)/80.-0.5
                          ysky4(i) = detycrpx  + ynew(k) &
                               - (srnew(k)*bix + crnew(k)*biy)/80.-0.5
                       end do
                       xsky4(0)=xsky4(4)
                       xsky4(5)=xsky4(1)
                       ysky4(0)=ysky4(4)
                       ysky4(5)=ysky4(1)
                       fk = sqrt(0.5*((xsky4(3)-xsky4(1))**2 + &
                                     (ysky4(3)-ysky4(1))**2))
                       pos=.false.
                       do i=0,nband-1
                         ! wght(i) = area(i)*transmission(i)*effgrad(i)* &
                         ! fexpext(iccd,k)
                         wgh = area(i)*fexpext(iccd,k)*fk**2
                         if (wgh > 0.0) then
                           wght(i)=wgh
                           pos=.true.
                         end if
                       end do 
                       if (pos) then

! Precalculation of resampling parameters to solve the speed issue. Alexis.
                         if(qalpha(k)) then
                            qalpha(k) = .false.
                            dx=xsky4(3)-xsky4(2)
                            dy=ysky4(3)-ysky4(2)
                            alpha = atan2(dy,dx)*180.d0/pi
                            if(alpha .lt. 0.d0) alpha=360.d0+alpha
                            alpha = alpha - 90.d0* (int(alpha)/90)
                               q45(k)=.false.
                            if(alpha .gt. 45.d0) then
                               q45(k)=.true.
                               alpha=90.d0 - alpha
                            endif
                            alpha=alpha*pi/180.d0
                            hs(k) = dsqrt(0.5d0)*dsin((pi/4.d0-alpha))
                            alcos(k) = dcos(alpha)
                            alsin(k) = dsin(alpha)
                         endif

                         call pix2map (xsky4,ysky4,wght,fk, &
                              naxis1, naxis2, nband, &
                              alsin(k),alcos(k),hs(k),q45(k))
                         !call pix2map (xsky4,ysky4,wght,fk, &
                         !     naxis1, naxis2, nband)
                       end if
                     end if  
                    end do

                 end if                  

                end if
             end do ! ipy
          end do ! ipyp
       end do ! ipy
     end if
    end do
   
    if (.not. usedetector) then
     if (fastmode) then
      do i=0,nband-1
        earr(i)%exp=0.0
      end do
      exptest=-1.0
     end if
      do ix=0,naxis1-1
        do iy=0,naxis2-1
         if (fastmode)  then
          do k=1,natt 
            if (exp0(k,ix,iy)>0.0) then
              do i=0,nband-1
                earr(i)%exp(ix,iy)=earr(i)%exp(ix,iy) + &
                                   exp1(i,k,ix,iy)/exp0(k,ix,iy)
              end do
            end if
          end do
         end if
          if (earr(0)%exp(ix,iy)>exptest) then
            exptest=earr(0)%exp(ix,iy)
            xcen=ix
            ycen=iy
          end if
        end do
      end do
    end if

    do i=0,nband-1 
      call cp_head (set1, setO(i), "EPIC EXPOSURE MAP", 1)
      call setAttribute(setO(i),"XCEN",xcen,"X pos. of opt. axis")
      call setAttribute(setO(i),"YCEN",ycen,"Y pos. of opt. axis")
      call setAttribute(setO(i),"ATT_SRC","AHF     ","Source of attitude data (AHF|RAF|THF)")
      call release(arr(i))
      call release(setO(i)) 
    end do

    if (nbb > 0) then
      call release (crawx)
      call release (crawy)
      call release (cye)
      call release (tabcoras)
    end if
    if (closeos) then 
! clean previous extension
!       call release (cox)
!       call release (cofs)
       call release (taboffs)
    end if

    call release (arr0)
    call release (set1)

    call release (set2)
    call release (set3)


  end subroutine eexpmap

  subroutine atthst (ranomrad, sdecnom, cdecnom, skypixsize, time, ra, de, pa&
       , ncoras, gti, ngti, ngti_ext, nccds, set, xnew, ynew, crnew, srnew,&
       & npnt, eps, winx0, winy0, windx, windy, nextmax)

    type(start_stopT), dimension(:) :: gti 
    type(EulerAnglesType) :: euler
    real(kind=double), parameter  :: radian = 57.29577951d0
    real(kind=double), dimension(:) :: xnew, ynew, crnew, srnew
!    real(kind=double), dimension(:,:), allocatable :: fexpext
    real(kind=double) :: eps, fexp1, fexp0, fexp00 
    !integer(kind=int32), parameter :: nattmax=10000
    integer(kind=int32), dimension(12) :: winx0, winy0, windx, windy
    integer(kind=int32) :: ngti_ext, npnt, npnt1, ipnt, ncoras, j, iext, &
                           i, i0, nccds, nxx, iccd, nextmax
    real(kind=double) :: ranomrad, sdecnom, cdecnom, skypixsize
    integer(kind=int32), dimension(:) :: ngti 
    integer(kind=int32), dimension(:), allocatable :: indx 
    real(kind=double), dimension(:), pointer :: time, ra, de, pa
    real(kind=double), dimension(:,:), pointer :: Bmatrix
    real(kind=double) :: rahelp, dechelp, relxb, relyb, rnb, rnew, sinr, cosr,&
         bor1, bor2, bor3, rn, relx, rely, relx2, rely2, diffPA, attPAxmm,&
         & attPArad, attdecrad, attrarad
    real(kind=double) :: tbeg, tend
    character(len=20) :: name
    character(len=6) :: xxname
    type(DataSetT) :: set
    type(TableT) :: tabcoras
    type(ColumnT) :: ctime,cfrac
    real(kind=single), dimension(:), pointer :: xxfrac
    real(kind=double), dimension(:), pointer :: xxtime

    allocate (indx(1:ncoras))

! 1) analyze attitude changes

    npnt=1
    do i0=0,ncoras-1
      if (dabs(ra(i0)).le.360d0 .and. &
          dabs(de(i0)).le. 90d0 .and. &
          dabs(pa(i0)).le.360d0) then
        indx(npnt)=i0
        exit
      end if
    end do
    i0=i0+1 
    if (i0.gt.1) call warning("NullValues", &
                 "NULL values in the attitude table were ignored") 

    do i=i0,ncoras-1

      if (dabs(ra(i)).le.360d0 .and. &
          dabs(de(i)).le. 90d0 .and. &
          dabs(pa(i)).le.360d0) then

       if (   (dabs(ra(i) - ra(indx(npnt)))) .gt. eps .or. &
            & (dabs(de(i) - de(indx(npnt)))) .gt. eps .or. &
            & (dabs(pa(i) - pa(indx(npnt)))) .gt. 200.0*eps) then
          npnt=npnt+1
          !if (npnt > nattmax) call error("BufferFull", &
          !              "More than 10000 attitude bins")
          indx(npnt)=i
       end if
      else
        call warning("NullValues", &
                     "NULL values in the attitude table were ignored")        
      end if
    end do

    allocate (fexpext(nextmax,npnt))    
    xxname = "EXPOSU"
   
! 2) merge with GTI to create the vector of factors

    do iccd=1,nccds

! set  default window ranges in case no exposure extension is present
       winx0(iccd)=0
       winy0(iccd)=0
       windx(iccd)=999    
       windy(iccd)=999 

       write(name,"(i2.2)") iccd  
       name=xxname//name(1:2)
       if (hasBlock(set,name)) then          
          tabcoras = table(set,name)

! get window ranges
          if (hasAttribute(tabcoras,"WINDOWX0")) then
            winx0(iccd)=real32Attribute(tabcoras,"WINDOWX0")
          else
            winx0(iccd)=0
          end if
          if (hasAttribute(tabcoras,"WINDOWY0")) then
            winy0(iccd)=real32Attribute(tabcoras,"WINDOWY0")
          else
            winy0(iccd)=0
          end if
          if (hasAttribute(tabcoras,"WINDOWDX")) then
            windx(iccd)=real32Attribute(tabcoras,"WINDOWDX")
          else
            windx(iccd)=999
          end if
          if (hasAttribute(tabcoras,"WINDOWDY")) then
            windy(iccd)=real32Attribute(tabcoras,"WINDOWDY")
          else
            windy(iccd)=999
          end if

          nxx = numberOfRows(tabcoras)
          ctime = column(tabcoras,"TIME", Read)    
          xxtime => real64Data(ctime)
          cfrac = column(tabcoras,"FRACEXP", Read)
          xxfrac => real32Data(cfrac)
       write(errstr,*) iccd, ' extension XX: N points is',nxx
       call message(AppMsg,VERBOSE)

       iext = min(ngti_ext, iccd) ! case of similar GTI for all chips

       do i=1,npnt
          fexpext(iccd,i)=0.
          do j=0,ngti(iext)-1
             if(gti(iext)%stop(j) .gt. time(indx(i))) then
                if (i == npnt) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(gti(iext)%stop(j),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxfrac, nxx)
                else if (i == 1) then
                   tbeg = gti(iext)%start(j)
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxfrac, nxx)
                else if (gti(iext)%start(j) .lt. time(indx(i+1))) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) + xxcti (tbeg, tend,&
                                     xxtime, xxfrac, nxx)                   
                end if
             end if
          end do
! exposure extension is accounted for by calling xxcti
!          write (errstr,*) i,iccd,fexpext(iccd,i)
!          call message(AppMsg,VERBOSE)
       end do

       else
          nxx=0
          call warning("NoExpoExt","Exposure extension not found")
       iext = min(ngti_ext, iccd) ! case of similar GTI for all chips
       do i=1,npnt                                      
          fexpext(iccd,i)=0.
          do j=0,ngti(iext)-1
             if(gti(iext)%stop(j) .gt. time(indx(i))) then
                if (i == npnt) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(gti(iext)%stop(j),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                else if (i == 1) then
                   tbeg = gti(iext)%start(j)
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                else if (gti(iext)%start(j) .lt. time(indx(i+1))) then
                   tbeg = max( time(indx(i)), gti(iext)%start(j) )
                   tend = max(min(gti(iext)%stop(j), time(indx(i+1))),tbeg)
                   fexpext(iccd,i) = fexpext(iccd,i) - tbeg + tend
                end if
             end if
          end do
! exposure extension is accounted for by calling xxcti
!          write (errstr,*) i,iccd,fexpext(iccd,i)
!          call message(AppMsg,VERBOSE)
       end do
       end if

       if (hasBlock(set,name)) then
          call release (ctime)
          call release (cfrac)
          call release (tabcoras)
       end if
    end do

! 2b) identify time bins with similar attitude    

     npnt1=npnt
     fexp0=0d0
     do ipnt=1,npnt
       fexp00=0d0
       do iccd=1,nccds
         fexp00=max(fexp00,fexpext(iccd,ipnt))
       end do
       fexp0=fexp0+fexp00
     end do 
     fexp0=fexp0*1d-4

     do ipnt=npnt,1,-1
       fexp1=0d0
       do iccd=1,nccds
         fexp1=max(fexp1,fexpext(iccd,ipnt))
       end do
       if (fexp1<fexp0) then
         npnt1=npnt1-1
         do j=ipnt,npnt1
           do iccd=1,nccds
             fexpext(iccd,j)=fexpext(iccd,j+1)
           end do          
           indx(j)=indx(j+1)
         end do
       end if
     end do

     do ipnt=npnt1,1,-1
       do i=ipnt-1,1,-1
         if ( (dabs(ra(indx(i)) - ra(indx(ipnt)))) < eps .and. &
              (dabs(de(indx(i)) - de(indx(ipnt)))) < eps .and. &
              (dabs(pa(indx(i)) - pa(indx(ipnt)))) < 200.0*eps) then
           do iccd=1,nccds
             fexpext(iccd,i)=fexpext(iccd,i)+fexpext(iccd,ipnt)
           end do          
           npnt1=npnt1-1
           do j=ipnt,npnt1
             do iccd=1,nccds
               fexpext(iccd,j)=fexpext(iccd,j+1)
             end do          
             indx(j)=indx(j+1)
           end do
           exit   
         end if
       end do
     end do
     npnt=npnt1

    write (errstr,*) 'number of att-histogram bins ', npnt
    call message(AppMsg,VERBOSE)

! 3) get transformation to sky coordinates

    do i=1,npnt
       
       attrarad = ra(indx(i))/radian
       attdecrad = de(indx(i))/radian
       attPArad = pa(indx(i))/radian

       attPAxmm = 4.712388980 - attPArad  
       call spher2tang (attrarad, attdecrad, ranomrad, sdecnom, cdecnom, &
            relx, rely)
    !--------------------------------------------------------------!
    !	do same 1deg away local north (help point) (SUBR) Ysc-axis !
    !--------------------------------------------------------------!
       call helppoint (attrarad, attdecrad, attPAxmm, & 
            rahelp, dechelp)
       call spher2tang (rahelp, dechelp, ranomrad , sdecnom, cdecnom, & 
            relx2, rely2)
    !-----------------------!     
    !	PA correction       ! 
    !-----------------------!
       diffPA = datan2(relx-relx2, rely2-rely)
       rn = sngl(4.71238898 - diffPA) !270 - diffpa; rn now in apa sense
       if (rn > 6.28319) rn = rn - 6.283185307
       if (rn < 0.0) rn = rn + 6.283185307

    !------------------------------------------!
    ! 	get boresight (SIAM) Bmatrix and angles         
    !------------------------------------------!
       call CAL_getBoresightMatrix(time(indx(i)), Bmatrix)
       call OAL_toEulerAngles(Bmatrix, euler)
        !-----------------------------------------------------------!
        !  convert to bor3, bor2, bor1  - for the moment, hardwire  !
        !-----------------------------------------------------------!
       bor3 = euler%phi
       bor2 = euler%theta
       bor1 = euler%psi
       call CAL_releaseMemory(Bmatrix)
       bor3 = dsin(bor3)
       bor2 = dsin(bor2)
       bor1 = dsin(bor1)
! test - no boresight
!        bor3 = 0.0 !dsin(pi)
!        bor2 = 0.0 !dsin(0.0)
!        bor1 = 0.0 !dsin(0.0)
        !---------------------------------------------!
        !   apply boresight correction (SUBR) to att  !
        !---------------------------------------------!
       sinr = dsin(dble(rn))
       cosr = dcos(dble(rn))
       call borxy (bor3, bor2, bor1, relx, rely, rn, relxb, relyb, rnb, &
            sinr, cosr)
       rnew = rnb * radian			!rnew in degrees in apa sense
       srnew(i) = sin(rnb)			!i.e. on sky anticlock (celN)
       crnew(i) = cos(rnb)			!xnew, ynew in x/y frame
       xnew(i) = relxb * radian/skypixsize	!all b/s corrected
       ynew(i) = relyb * radian/skypixsize 	!
   end do

  end subroutine atthst

!---------------------------------------------------------------------
!
  function xxcti(tbeg, tend, xxtime, xxfrac, nxx)
!
!---------------------------------------------------------------------
! eXposure extension Corrected Time Interval
    real(kind=double) :: tbeg, tend, sum, xxcti
    integer(kind=int32) :: nxx, nin, i
    real(kind=single), dimension(:), pointer :: xxfrac
    real(kind=double), dimension(:), pointer :: xxtime
    

    xxcti = tend - tbeg
       
    if (nxx == 0) return

    nin = 0
    sum = 0.
    do i = 0, nxx-1
       if(xxtime(i) .ge. tbeg .and. xxtime(i) .le. tend) then
          nin=nin+1
          sum=sum+xxfrac(i)
       end if
    end do

    if (nin == 0) return
       
    xxcti = xxcti * sum / nin

  end function xxcti

!---------------------------------------------------------------------
  real function get_BadRadius(ibbc)
    !Event pattern selection of ibbc (higher value includes lower values)
    !  5 4 3 4 5
    !  4 2 1 2 4
    !  3 1 0 1 3
    !  4 2 1 2 4
    !  5 4 3 4 5
    !
    ! For PN valid grades up to 2. For MOS up to 5.

    real(kind=single), dimension (0:5) :: radius
    integer(kind=int32), intent(inout) :: ibbc

    if(ibbc .gt. 5) then
       write(errstr,*) 'Not valid grade for cleaning! Must be < 5'
       call message(AppMsg,Verbose)
       ibbc = 5
    end if

    radius(0)=0.
    radius(1)=1.0
    radius(2)=1.42
    radius(3)=2.0
    radius(4)=2.24
    radius(5)=2.83


    get_BadRadius = radius(ibbc)+0.01

  end function get_BadRadius


!---------------------------------------------------------------------
  logical function wmode (iccd, nsubmode, ipx, ipy)

  integer, intent(in) :: iccd, nsubmode
  integer(kind=int16), intent(in) :: ipx, ipy

  wmode = .true.

  if (nsubmode >= 15 .and. ipy <= 15) wmode=.false. 

  if (nsubmode == 17) then
    if (ipy < 101) wmode = .false.
  else if (nsubmode == 18) then
    if (iccd /= 4 .or. iccd == 4 .and. ipy < 137) wmode = .false.
  else if (nsubmode == 2 .or. nsubmode == 4) then
    if (iccd == 1 .and. &
       (ipx < 251 .or. ipx > 350 .or. & 
        ipy < 251 .or. ipy > 350)) wmode = .false.
  else if (nsubmode == 3 .or. nsubmode == 5) then
    if (iccd == 1 .and. &
       (ipx < 151 .or. ipx > 450 .or. & 
        ipy < 151 .or. ipy > 450)) wmode = .false.
  end if

  end function wmode 

!---------------------------------------------------------------------
!
!	*** SUBROUTINES from attcalc
!
!---------------------------------------------------------------------
    subroutine spher2tang (attra, attdec, ranom, sdecnom, cdecnom, relx, &
                           rely)
    implicit none
    real(kind=double)                     :: attra, attdec, sdecnom, &
                                        cdecnom, relx, rely, &
                                        ranom, diffra, dsinra, &
                                        dcosra, dsindec, dcosdec, &
                                        d1s, dh, d1c, dranew, d2s, d2c, &
                                        ddecnew

    diffra = attra - ranom
    dsinra = dsin(diffra)
    dcosra = dcos(diffra)
    dsindec = dsin(attdec)
    dcosdec = dcos(attdec)

!	residual alpha dranew
    d1s = dsinra * dcosdec
    dh = - dcosdec * dcosra
    d1c = sdecnom*dsindec - dh*cdecnom
    dranew = datan2(d1s,d1c)

! 	residual delta ddecnew
    d2s = dh*sdecnom + dsindec*cdecnom
    d2c = dsqrt(1.0d0 - d2s*d2s)
    ddecnew = datan2(d2s,d2c)

! 	convert to x,y in NS-system
    relx = - dranew
    rely =  - ddecnew
    end subroutine spher2tang
  !---------------------------------------------------------------------
    subroutine helppoint (attra, attdec, attPA, &
                           rahelp, dechelp)

    implicit none
    real(kind=single), parameter        :: sin1deg = 1.74524064E-2, &
         cos1deg = 9.99847695E-1
    real(kind=double)           :: attra, attdec, attPA, &
                              rasin, decsin, PAsin, racos, deccos, &
                              PAcos, xhelp, yhelp, &
                              zhelp, rhelp, rahelp, &
                              dechelp

!	calc sines and cosines
      rasin = dsin(attra)
      racos = dcos(attra)
      decsin = dsin(attdec)
      deccos = dcos(attdec)
      PAsin = dsin(attPA)
      PAcos = dcos(attPA)

!	cartesian co-ords of help point
       xhelp = (sin1deg*(PAcos*racos*decsin - PAsin*rasin)) + &
                        (cos1deg*racos*deccos)
       yhelp = (sin1deg*(PAsin*racos + PAcos*rasin*decsin)) + &
                        (cos1deg*rasin*deccos)
       
       zhelp = - sin1deg*PAcos*deccos + cos1deg*decsin

!	length squared of help vector
      rhelp = xhelp*xhelp + yhelp*yhelp +zhelp*zhelp

      if (rhelp < 0.d0) then
!	error
        call warning("undefinedHelpVector","Help vector has undefined length - taking aspect solution as help point to contine")
        rahelp = attra
        dechelp = attdec
      else

!	spherical co-ords of help point
        rhelp = dsqrt(rhelp)
        rahelp = datan2 (yhelp, xhelp)
          if (rahelp < 0.d0) rahelp = rahelp + (2*3.1415926536)
        dechelp = dasin(zhelp/rhelp)
      endif

    end subroutine helppoint
  !---------------------------------------------------------------------
    subroutine borxy (bor3, bor2, bor1, relx, rely, rn, relxb, &
                      relyb, rnb, sinr, cosr)

    implicit none
    real(kind=double)       :: bor3, bor2, bor1, relx, rely, rn, relxb, &
                            relyb, rnb, sinr, cosr, xbor, ybor

    xbor = 0.0
    ybor = 0.0
    xbor = (cosr * bor3) + (sinr * bor2) !in x/y frame
    ybor = (sinr * bor3) - (cosr * bor2) !in x/y frame

!       shift attitude vector by (xbor,ybor)

    relxb = relx + xbor
    relyb = - rely + ybor                !minus to put rely in x/y frame

!       correct "corrected Pos.Angle" also by fibor
    rnb = rn + bor1

    end subroutine borxy

!-----------------------------------------------------------------
 subroutine merge_gti (lgti, ngti, fstart, fstop, irange, itfilt, itfilt1)

 real(kind=double), dimension(0:5000), intent(IN) :: fstart, fstop 
 integer, intent(IN) :: irange, itfilt
 integer, intent(INOUT) :: ngti
 
 integer :: j,k, n, m, ntest, mtest, itfilt1
 real(kind=double) :: aux
 real(kind=double), dimension(2*(ngti+irange)) :: tbuff
! real(kind=double), dimension(0:ngti+irange-1), target :: tstart, tstop 
 logical, dimension(2*(ngti+irange)) :: type
 logical :: bux
 
 type(start_stopT) :: lgti

!   itfilt1=itfilt+ngti-1

   k=1
   do j=0,ngti-1
     tbuff(k)=lgti%start(j)
     type(k)=.true. 
     k=k+1
     tbuff(k)=lgti%stop(j)
     type(k)=.false.
     k=k+1
   end do 
   do j=0,irange-1
     tbuff(k)=fstart(j)
     type(k)=.true.
     k=k+1
     tbuff(k)=fstop(j)
     type(k)=.false.
     k=k+1
   end do

   do n=1,k-1
     do m=n+1,k-1
       if (tbuff(n)>tbuff(m)) then
         aux=tbuff(n) 
         tbuff(n)=tbuff(m)
         tbuff(m)=aux
         bux=type(n)
         type(n)=type(m)
         type(m)=bux
       end if
     end do
   end do 


   if (ngti>0) then
     deallocate (lgti%start)
     deallocate (lgti%stop)
   end if 
   allocate (lgti%start(0:ngti+irange-1)) 
   allocate (lgti%stop(0:ngti+irange-1)) 

   ntest=0
   j=0 
   do n=1,k-1
     mtest=ntest
     if (type(n)) then
       ntest=ntest+1
     else
       ntest=ntest-1
     end if
     if (ntest == itfilt1+1 .and. mtest == itfilt1) then
       lgti%start(j) = tbuff(n)
!       tstart(j) = tbuff(n)
     end if
     if (ntest == itfilt1 .and. mtest == itfilt1+1) then 
       lgti%stop(j) = tbuff(n)
!       tstop(j) = tbuff(n)
       j=j+1
     end if
   end do

   ngti = j

!   gti(i)%start => tstart
!   gti(i)%stop => tstop
  
 end subroutine merge_gti
  
end module eexpmap_mod  
  




