module pix2map_mod

  use cal
  use oal
  use param  
  use dal 
 
  implicit none 

  type :: ExpoT
     real(kind=single), dimension(:,:), pointer :: exp
  end type
  type(ExpoT), dimension(:), allocatable :: earr
   
  public pix2map, cross, icross, linesec, sqseca, sqsec
!, sind, cosd, tand

  

contains 


! The purpose of this program is to perform the pixelization in the sky
! coordinates of the detector image.

subroutine pix2map (xsky,ysky,wght0,fac0, nm1, nm2, nband, alsin, alcos&
     & , hs, q45s)

  implicit none
! input/output parameters
  real(kind=single), dimension(:), intent(in) :: wght0
  real(kind=single), intent(in) :: fac0
  real(kind=single), intent(in), dimension(0:5) :: xsky, ysky
  real(kind=single), intent(in) :: alsin, alcos, hs
!  real(kind=single), dimension(:,:), intent(inout) :: imh
  integer(kind=int32), intent(in) :: nm1, nm2, nband
  
  real(kind=single) :: fac,alpha,alpha1
  integer(kind=int32), dimension(2) :: indx, indy
  integer(kind=int32) ::  i,j,k,nin,inn
  integer(kind=int32) ::  isky,isky0,isky1,jsky,jsky0,jsky1,iskyx,jskyy
  logical :: in,inv,tran,tranx, q45z, q45s
  real(kind=single) :: x1,x2,y1,y2,vp
  real(kind=single) :: dx,dy
  real(kind=single) :: wfac,xcntr,ycntr,d1,d2,vp1,sum
  real(kind=single), dimension(0:nband-1) :: wght 
  real(kind=single), dimension(4) :: alphas, alphaz

  fac=fac0
  wght=wght0/fac/fac
  q45z = .not. q45s


  ! find min and max values
  indx(1)=1
  indx(2)=1
  indy(1)=1
  indy(2)=1
  do i=2,4
     if(xsky(i) .lt. xsky(indx(1))) indx(1) = i
     if(xsky(i) .gt. xsky(indx(2))) indx(2) = i
     if(ysky(i) .lt. ysky(indy(1))) indy(1) = i
     if(ysky(i) .gt. ysky(indy(2))) indy(2) = i
  end do

!      print*,xsky(indx(1)),xsky(indx(2)), ysky(indy(1)), ysky(indy(2))
! we will treat all the corners of the pixels, so we have simply
  isky0=xsky(indx(1))
  isky1=xsky(indx(2))
  jsky0=ysky(indy(1))
  jsky1=ysky(indy(2))
  !      print*,isky0,isky1,jsky0,jsky1
  ! when porting this program furthermore, here one can save by exiting on 
  ! negative values of isky, jsky
  ! main loop
  sum=0. ! this is debugging variable
  do isky=isky0,isky1
     do jsky=jsky0,jsky1
! step 1: analyze the corners, find the right case
        if(isky .gt. 0 .and. jsky .gt. 0 .and. isky .le. nm1 .and. &
             jsky .le. nm2) then
!            print*,'processing',isky,jsky
           nin=0
           inv=.false.
           do j=1,4
              in=.true.
              do i=1,4
                 x1=0.5*sign(1+(-1)**(i+1),(-1)**((i+1)/2))
                 y1=0.5*sign(1+(-1)**i,(-1)**((i+2)/2))
                 iskyx=isky+0.5*(1+(-1)**((i+1)/2))
                 jskyy=jsky + 0.5*(1+(-1)**((i+2)/2))
                 x2=xsky(j)-iskyx+x1
                 y2=ysky(j)-jskyy+y1
                 vp=x1*y2-x2*y1
                 if(vp .ge. 0.) in=.false.

              end do
              if(in) then
                 nin=nin+1
                 inn=j
              end if
           end do
! inverse case
           if(nin .eq. 2 .or. nin .eq. 3) then
              inv=.true.
              nin=0
              do j=1,4
                 in=.true.
                 iskyx=isky+0.5*(1+(-1)**((j+1)/2))
                 jskyy=jsky + 0.5*(1+(-1)**((j+2)/2))
                 do i=1,4
                    x1=xsky(i)-xsky(i-1)
                    y1=ysky(i)-ysky(i-1)
                    x2=iskyx-xsky(i-1)
                    y2=jskyy-ysky(i-1)
                    vp=x1*y2-x2*y1
!                print*,j,vp,x1,y1,x2,y2
                    if(vp .ge. 0.) in=.false.
                 end do
                 if(in) then
                    nin=nin+1
                    inn=j
                 end if
              end do
           end if
! impossible case
           if(nin .eq. 2 .or. nin .eq. 3) then
              print*,'This is impossible'
              stop
           end if

! step 2: find distances to the transformed pixels edges
           if(nin .eq. 4) then
!            print*,'n=4: pixel is totally within',isky,jsky
             do k=0,nband-1
               earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+ wght(k)*fac**2
             end do
!             sum=sum+wght*1.*fac**2
           elseif(inv) then
              xcntr=0.5*(xsky(1)+xsky(3))
              ycntr=0.5*(ysky(1)+ysky(3))
              if(nin .eq. 1) then
! account for the corner
                 j=inn
                 x1=0.5*(1+(-1)**((j+1)/2)) - 0.5*(1+(-1)**((j)/2))
                 y1=0.5*(1+(-1)**((j+2)/2)) - 0.5*(1+(-1)**((j+1)/2))
                 x2=xcntr - 0.5*(1+(-1)**((j)/2)) - isky
                 y2=ycntr -  0.5*(1+(-1)**((j+1)/2)) - jsky
                 vp=y2*x1 - x2*y1
                 d1=vp/sqrt(x1*x1+y1*y1)
! this d contains both right sign and module, thanks to VectorProduct
                 x1=0.5*(1+(-1)**((j+2)/2)) - 0.5*(1+(-1)**((j+1)/2))
                 y1=0.5*(1+(-1)**((j+3)/2)) - 0.5*(1+(-1)**((j+2)/2))
                 x2=xcntr - 0.5*(1+(-1)**((j+1)/2)) - isky
                 y2=ycntr -  0.5*(1+(-1)**((j+2)/2)) - jsky
                 vp=y2*x1 - x2*y1
                 d2=vp/sqrt(x1*x1+y1*y1)
                 d1=d1/fac
                 d2=d2/fac
!              print*,'making inv',j,d1,d2,sqseca(d1,d2,alpha)
                 wfac=sqseca(d1,d2,alsin,alcos,hs,q45z)*fac*fac
                 do k=0,nband-1
                   earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+ &
                                   wght(k)*wfac
                 end do
!                 sum=sum+wght*sqseca(d1,d2,alphaz(j))*fac*fac
              else                ! nin .eq. 0
! consider only 1 side of interest, given the combination possible in such
! an arrangement
                 tranx=.true.
                 do j=1,4
                    x1=0.5*(1+(-1)**((j+1)/2)) - 0.5*(1+(-1)**((j)/2))
                    y1=0.5*(1+(-1)**((j+2)/2)) - 0.5*(1+(-1)**((j+1)/2))
                    x2=xcntr-isky-0.5*(1+(-1)**((j)/2))
                    y2=ycntr-jsky-0.5*(1+(-1)**((j+1)/2))
                    vp1= x2*y1 - y2*x1
                    if(vp1 .le. 0) tranx=.false.
                 end do
                 if(tranx) then
                   do k=0,nband-1
                     earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+wght(k) &
                                           *fac*fac
                   end do
                 end if 
!                 if(tranx) sum=sum+wght*1.*fac*fac

                 do j=1,4
                    tran=icross(xsky,ysky,isky,jsky,j)
                    if(tran) then
                       ! treat the transition
                       x1=0.5*(1+(-1)**((j+1)/2)) - 0.5*(1+(-1)**((j)/2))
                       y1=0.5*(1+(-1)**((j+2)/2)) - 0.5*(1+(-1)**((j+1)/2))
                       x2=xcntr-isky-0.5*(1+(-1)**((j)/2))
                       y2=ycntr-jsky-0.5*(1+(-1)**((j+1)/2))
                       vp=y2*x1 - x2*y1
                       d1=vp/sqrt(x1*x1+y1*y1)
                       d1=d1/fac
                       if(d1 .lt. 0.) then
                          d1=-d1
!                    print*,'in case i0 transection is subtracted'
                          wfac=linesec(d1,alsin,alcos,hs)*fac**2
                          do k=0,nband-1 
                            earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)- &
                               wght(k)*wfac
                          end do
!                          sum=sum-wght*linesec(d1,alphas(j))*fac**2
                       else
!                    print*,'in case i0 transection is added'
                          wfac=linesec(d1,alsin,alcos,hs)*fac**2
                          do k=0,nband-1
                            earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+ &
                              wght(k)*wfac
                          end do
!                          sum=sum+wght*linesec(d1,alphas(j))*fac**2
                       end if
!                  print*,j,d1,alphas(j),(linesec(d1,alphas(j)))
                    end if
                 end do

               end if
            else
               if(nin .eq. 1) then
!              print*,'processing case n=1'
! account for the corner
!              if((abs(mod(alpha,90.)) .lt. 0.01) then
                  x1=xsky(inn)-xsky(inn-1)
                  y1=ysky(inn)-ysky(inn-1)
                  x2=isky+0.5-xsky(inn-1)
                  y2=jsky+0.5-ysky(inn-1)
                  vp=y2*x1 - x2*y1
                  d1=vp/sqrt(x1*x1+y1*y1)
! this d contains both right sign and module, thanks to VectorProduct
                  x1=xsky(inn+1)-xsky(inn)
                  y1=ysky(inn+1)-ysky(inn)
                  x2=isky+0.5-xsky(inn)
                  y2=jsky+0.5-ysky(inn)
                  vp=y2*x1 - x2*y1
                  d2=vp/sqrt(x1*x1+y1*y1)
                  wfac=sqseca(d1,d2,alsin,alcos,hs,q45s)
                  do k=0,nband-1
                    earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky) + &
                       wght(k)*wfac
                  end do
!                  sum=sum+wght*sqseca(d1,d2,alphas(inn))
!              print*,d1,d2,alphas(inn),(sqseca(d1,d2,alphas(inn))), &
! (sqseca(d2,d1,alphas(inn)))

! account for the 2 sides left 
!              print*,'here is loop for the left two sides:'
!              print*,(max(1,mod(inn+2,4))),(inn-1),(inn+2),4
                  do j=max(1,mod(inn+2,4)),inn-1
                     if(cross(xsky(j),ysky(j),xsky(j-1),ysky(j-1),isky,jsky))&
                          & then
! transection is found
                        x1=xsky(j)-xsky(j-1)
                        y1=ysky(j)-ysky(j-1)
                        x2=isky+0.5-xsky(j-1)
                        y2=jsky+0.5-ysky(j-1)
                        vp=y2*x1 - x2*y1
! we need two invertions of a sign: here and in -linesec
                        d1=-vp/sqrt(x1*x1+y1*y1)
                        wfac=linesec(d1,alsin,alcos,hs)
                        do k=0,nband-1
                          earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)- &
                             wght(k)*wfac
                        end do
!                        sum=sum-wght*linesec(d1,alphas(j))
!                  print*,j,' left sides ',d1,alphas(j),(linesec(d1,alphas(j)))
                     end if
                  end do
! repeat the same 
                  do j=inn+2,4
                     if(cross(xsky(j),ysky(j),xsky(j-1),ysky(j-1),isky,jsky))&
                          & then
! transection is found
                        x1=xsky(j)-xsky(j-1)
                        y1=ysky(j)-ysky(j-1)
                        x2=isky+0.5-xsky(j-1)
                        y2=jsky+0.5-ysky(j-1)
                        vp=y2*x1 - x2*y1
! we need two invertions of a sign: here and in -linesec
                        d1=-vp/sqrt(x1*x1+y1*y1)
                        wfac=linesec(d1,alsin,alcos,hs) 
                        do k=0,nband-1
                          earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)- &
                            wght(k)*wfac
                        end do
!                        sum=sum-wght*linesec(d1,alphas(j))
!                  print*,j,' left sides ',d1,alphas(j),(linesec(d1,alphas(j)))
                     end if
                  end do

               else                ! nin .eq. 0
! consider all 4 sides
                  tranx=.true.
                  do j=1,4
                     x1=xsky(j)-xsky(j-1)
                     y1=ysky(j)-ysky(j-1)
                     x2=isky+0.5-xsky(j-1)
                     y2=jsky+0.5-ysky(j-1)
                     vp1= x2*y1 - y2*x1
                     if(vp1 .le. 0) tranx=.false.
                  end do
                  if(tranx) then
                    do k=0,nband-1
                      earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+wght(k)
                    end do
                  end if
!                  if(tranx) sum=sum+wght*1.
!              if(tranx) print*,isky,jsky,' pixel is inside n=0'

                  do j=1,4
                     x1=xsky(j)-xsky(j-1)
                     y1=ysky(j)-ysky(j-1)
                     tran=.false.
                     tran=cross(xsky(j),ysky(j),xsky(j-1),ysky(j-1),isky,jsky)
                     if(tran) then
! treat the transition
                        x2=isky+0.5-xsky(j-1)
                        y2=jsky+0.5-ysky(j-1)
                        vp=y2*x1 - x2*y1
                        d1=vp/sqrt(x1*x1+y1*y1)
                        if(d1 .lt. 0.) then
                           d1=-d1
!                  print*,'in case 0 transection is subtracted'
                           wfac=linesec(d1,alsin,alcos,hs)
                           do k=0,nband-1
                             earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)- &
                               wght(k)*wfac
                           end do
!                           sum=sum-wght*linesec(d1,alphas(j))
                        else
!                  print*,'in case 0 transection is added'
                          wfac=linesec(d1,alsin,alcos,hs)
                          do k=0,nband-1 
                            earr(k)%exp(isky,jsky)=earr(k)%exp(isky,jsky)+ &
                               wght(k)*wfac
                          end do
!                           sum=sum+wght*linesec(d1,alphas(j))
                        end if
!                  print*,j,d1,alphas(j),(linesec(d1,alphas(j)))
                     end if
                  end do
! the pixel can be just inside. so if pixels center is inside add 1.
               end if
            end if

        end if
      end do
    end do
!    sum = sum/wght0
!    if ( abs(sum-1.) .gt. 0.05) print*,sum,fac, alpha,nin,inv,tranx,tran

end subroutine pix2map




! real function sind(x0)
!   real(kind=single) :: x0,x
!   real(kind=single), parameter  :: radian = 57.29577951E0    
!   
!   x = x0 / radian
!   sind = sin (x)
! end function sind
! 
! real function cosd(x0)
!   real(kind=single) :: x0,x
!   real(kind=single), parameter  :: radian = 57.29577951E0    
!   
!   x = x0 / radian
!   cosd = cos (x)
! end function cosd
! 
! real function tand(x0)
!   real(kind=single) :: x0,x
!   real(kind=single), parameter  :: radian = 57.29577951E0    
!   
!   x = x0 / radian
!   tand = tan (x)
! end function tand


logical function cross(x1,y1,x2,y2,isky,jsky)
  real(kind=single), intent(in) :: x1,y1,x2,y2
  integer(kind=int32), intent(in) :: isky,jsky
  integer(kind=int32) :: i1,j1,i2,j2,i
  real(kind=single) :: vx1,vx2,vy1,vy2,vx3,vy3,vp1,vp2
  logical :: t1,t2,t3

  cross=.false.
  do i=1,4
     i1 = isky+0.5*(1+(-1)**((i+1)/2))
     j1 = jsky + 0.5*(1+(-1)**((i+2)/2))
     i2 = isky+0.5*(1+(-1)**((i+2)/2))
     j2 = jsky + 0.5*(1+(-1)**((i+3)/2))

     vx1=x2-x1
     vy1=y2-y1
     vx2=i1-x1
     vy2=j1-y1
     vx3=i2-x1
     vy3=j2-y1
        
     vp1=vy2*vx1 - vx2*vy1
     vp2=vy3*vx1 - vx3*vy1
     t1=((vp1*vp2) .lt. 0.)
     t2=(vp2 .eq. 0. .and. vp1 .lt. 0.)
     t3=(vp1 .eq. 0. .and. vp2 .lt. 0.)
     if(t1 .or. t2 .or. t3) then
        vx1=i2-i1
        vy1=j2-j1
        vx2=x1-i1
        vy2=y1-j1
        vx3=x2-i1
        vy3=y2-j1
        
        vp1=vy2*vx1 - vx2*vy1
        vp2=vy3*vx1 - vx3*vy1
        t1=((vp1*vp2) .lt. 0.)
        t2=(vp2 .eq. 0. .and. vp1 .lt. 0.)
        t3=(vp1 .eq. 0. .and. vp2 .lt. 0.)
        if(t1 .or. t2 .or. t3) cross=.true.
        
     end if
          
  end do
end function cross


logical function icross(xsky,ysky,isky,jsky,j)
  implicit none
  real(kind=single), intent(in), dimension(0:5) :: xsky, ysky
  integer(kind=int32), intent(in) :: isky,jsky,j
  real(kind=single) :: x1,y1,x2,y2
  real(kind=single) :: i1,j1,i2,j2
  integer(kind=int32) :: i
  real(kind=single) ::  vx1,vx2,vy1,vy2,vx3,vy3,vp1,vp2
  logical :: t1,t2,t3

  icross=.false.
  x1=isky + 0.5*(1+(-1)**((j)/2))
  y1=jsky + 0.5*(1+(-1)**((j+1)/2))
  x2=isky + 0.5*(1+(-1)**((j+1)/2))
  y2=jsky + 0.5*(1+(-1)**((j+2)/2))
      
  do i=1,4
     i1 = xsky(i-1)
     j1 = ysky(i-1)
     i2 = xsky(i)
     j2 = ysky(i)
     
     vx1=x2-x1
     vy1=y2-y1
     vx2=i1-x1
     vy2=j1-y1
     vx3=i2-x1
     vy3=j2-y1
        
     vp1=vy2*vx1 - vx2*vy1
     vp2=vy3*vx1 - vx3*vy1
     t1=((vp1*vp2) .lt. 0.)
     t2=(vp2 .eq. 0. .and. vp1 .lt. 0.)
     t3=(vp1 .eq. 0. .and. vp2 .lt. 0.)
     if(t1 .or. t2 .or. t3) then
        vx1=i2-i1
        vy1=j2-j1
        vx2=x1-i1
        vy2=y1-j1
        vx3=x2-i1
        vy3=y2-j1
        
        vp1=vy2*vx1 - vx2*vy1
        vp2=vy3*vx1 - vx3*vy1
        t1=((vp1*vp2) .lt. 0.)
        t2=(vp2 .eq. 0. .and. vp1 .lt. 0.)
        t3=(vp1 .eq. 0. .and. vp2 .lt. 0.)
        if(t1 .or. t2 .or. t3) icross=.true.
        
     end if
     
  end do
  
end function icross

real function linesec(d0,alsin,alcos,hs)
  implicit none
!.. d distance between the crossing line and analyzed pixel center, in pixels
!.. alpha angle between the line and X-coordinate in degrees
!   -0.5cos(alpha) < d < 0.5cos(alpha) 
!   negative values correspond to pixel center being inside 
  real(kind=single) :: d,d0,hl,hs,w,alsin,alcos
  logical :: inv
      
  linesec=0.
  d=d0
!  alpha=alpha0
!  if(alpha0 .lt. 0.) alpha=360+alpha0
!  alpha = alpha - 90.* (int(alpha)/90)

! Case of detector coordinates.
  if(alsin .lt. 0.01) then
     linesec=0.5-d
     if(linesec .lt. 0.) linesec=0.
     if(linesec .gt. 1.) linesec=1.
     return
  end if


!  w=45.-alpha
!  if(w .lt. 0.) w=alpha-45.
!  hs = sqrt(0.5)*sind(w)
  hl = 0.5*(alsin + alcos)

      
  if (d .ge. hl) return
  if (d .le. (-hl)) then
     linesec=1.
     return
  end if
  inv=.false.
  if (d .lt. 0) then
     inv = .true.
     d=-d
  end if

!  if (alpha .gt. 45.) alpha=90.-alpha

! Triangle case
  if(d .ge. hs) then
!     linesec=0.5*(hl-d)**2 / (hl-hs)/cosd(alpha)
     linesec=0.5*(hl-d)**2 / (hl-hs)/alcos
  else
! Trapezium case
!     linesec=0.5 - d/cosd(alpha)
     linesec=0.5 - d/alcos
  end if


  if(inv) then
     linesec=1. - linesec
  end if

end function linesec


real function sqseca(d10,d20,alsin,alcos,hs,q45)
! Meaning of alpha is important. It is an angle between Y and d2 (or X and d1)
  implicit none
  real(kind=single) d10,d20,d1,d2,fac1,fac2,alsin,alcos,hs
  logical :: q45

  sqseca=0.
!  alpha=alpha0
!  if(alpha0 .lt. 0.) alpha=360+alpha0
!  alpha = alpha - 90.* int(alpha/90.)
! Special case of detector coordinates.
!  if(alpha .lt. 0.01 .or. alpha .gt. 89.99) then
  if(alsin .lt. 0.01 .or. alcos .lt. 0.01) then
     fac1=0.5-d10
     fac2=0.5-d20
     if(d10 .le. -0.5) fac1=1.
     if(d10 .gt. 0.5) fac1=0.
     if(d20 .le. -0.5) fac2=1.
     if(d20 .gt. 0.5) fac2=0.
     sqseca=fac1*fac2
     return
  end if

  d1=abs(d10)
  d2=abs(d20)
  if((d10 .ge. 0. .and. d20 .ge. 0.) .or. (d10 .le. 0. .and. d20 .le.&
       & 0.)) then
     sqseca=sqsec(d1,d2,alsin,alcos,hs,q45)
  else                                
     sqseca=sqsec(d2,d1,alsin,alcos,hs,q45)
  end if
  
  if(d10 .ge. 0. .and. d20 .ge. 0.) return
  
  if(d10 .le. 0. .and. d20 .le. 0.) then
     sqseca = 1.-linesec(d1,alsin,alcos,hs)-linesec(d2,alsin,alcos,hs)+sqseca
  elseif(d10 .gt. 0) then
     sqseca = linesec(d1,alsin,alcos,hs)-sqseca
  else                                 
     sqseca = linesec(d2,alsin,alcos,hs)-sqseca
  end if

end function sqseca


real function sqsec(d10,d20,alsin,alcos,hs,q45)
  implicit none
! this program does not check anything: all the checks are done in SQSECA
!.. d distance between the OOTE-line and analyzed pixel center, in pixels
!.. alpha angle between OOTE-line and X-coordinate in degrees
!   0 < d < 0.5cos(alpha) 
!   negative values correspond to pixel center being inside 
  real(kind=single) :: hl,hs,w,d1,d2,d10,d20,alsin,alcos
  real(kind=single) :: a
  logical :: q45

  d1=d10
  d2=d20
  sqsec=0.
!  alpha=alpha0
  
  hl = 0.5*(alsin+alcos)
  if (d1 .lt. 0. .or. d2 .lt. 0.) return
  if (d1 .ge. hl .or. d2 .ge. hl) return
!  w=45.-alpha
!  if(w .lt. 0.) w=alpha-45.
!  hs = sqrt(0.5)*sind(w)
  
  if (q45) then
!     alpha=90.-alpha
     d1=d20
     d2=d10
  end if

  if(d1 .ge. hs) then
     a=(hl-d2)*alsin/alcos - (d1-hs)
     if(a .lt. 0) return
     sqsec=0.5*a*a*alcos/alsin
  else
     if(((hl-d2)*alcos) .lt. ((hs-d1)*alsin)) then
        sqsec=linesec(d2,alsin,alcos,hs)
! this case actually means that the corner is not inside.
        return
     else
        a=(hl-d2)
        sqsec=0.5*a*a*alsin/alcos
        sqsec=sqsec+0.5*(hs-d1)*(2.*(hl-d2)-(hs-d1)*alsin/alcos)
     end if
  end if

end function sqsec



end module pix2map_mod
